import type { PeerId } from '@libp2p/interface-peer-id';
import type { AddressSorter } from '@libp2p/interface-peer-store';
import type { Connection } from '@libp2p/interface-connection';
import { EventEmitter } from '@libp2p/interfaces/events';
import type { Startable } from '@libp2p/interfaces/startable';
import type { Components } from '../components.js';
import type { RelayReservationManagerConfig } from './index.js';
import { PeerList } from '@libp2p/peer-collections';
/**
 * CircuitServiceInit initializes the circuit service using values
 * from the provided config and an @type{AddressSorter}.
 */
export interface CircuitServiceInit extends RelayReservationManagerConfig {
    /**
     * Allows prioritizing addresses from the peerstore for dialing. The
     * default behavior is to prioritise public addresses.
     */
    addressSorter?: AddressSorter;
    /**
     * A callback to invoke when an error occurs in the circuit service.
     */
    onError?: (error: Error) => void;
}
export interface RelayReservationManagerEvents {
    'relay:reservation': CustomEvent;
}
/**
 * ReservationManager automatically makes a circuit v2 reservation on any connected
 * peers that support the circuit v2 HOP protocol.
 */
export declare class RelayReservationManager extends EventEmitter<RelayReservationManagerEvents> implements Startable {
    private readonly components;
    private readonly addressSorter;
    private readonly maxReservations;
    private readonly relays;
    private readonly reservationMap;
    private readonly onError;
    private started;
    constructor(components: Components, init: CircuitServiceInit);
    isStarted(): boolean;
    start(): void;
    stop(): Promise<void>;
    /**
     * Check if a peer supports the relay protocol.
     * If the protocol is not supported, check if it was supported before and remove it as a listen relay.
     * If the protocol is supported, check if the peer supports **HOP** and add it as a listener if
     * inside the threshold.
     */
    _onProtocolChange({ peerId, protocols }: {
        peerId: PeerId;
        protocols: string[];
    }): Promise<void>;
    /**
     * Handle case when peer connects. If we already have the peer in the protobook,
     * we treat this event as an `onProtocolChange`.
     */
    _onPeerConnect({ detail: connection }: CustomEvent<Connection>): void;
    /**
     * Peer disconnects
     */
    _onPeerDisconnected(evt: CustomEvent<Connection>): void;
    /**
     * Attempt to listen on the given relay connection
     */
    _addListenRelay(connection: Connection, peerId: PeerId): Promise<void>;
    /**
     * Remove listen relay
     */
    _removeListenRelay(PeerId: PeerId): Promise<void>;
    /**
     * Try to listen on available hop relay connections.
     * The following order will happen while we do not have enough relays.
     * 1. Check the metadata store for known relays, try to listen on the ones we are already connected.
     * 2. Dial and try to listen on the peers we know that support hop but are not connected.
     * 3. Search the network.
     */
    _listenOnAvailableHopRelays(peersToIgnore?: PeerList): Promise<void>;
    _tryToListenOnRelay(peerId: PeerId): Promise<void>;
    private readonly createOrRefreshReservation;
}
//# sourceMappingURL=client.d.ts.map