import * as CircuitV2 from './pb/index.js';
import { ReservationStore } from './reservation-store.js';
import { logger } from '@libp2p/logger';
import createError from 'err-code';
import * as mafmt from '@multiformats/mafmt';
import { multiaddr } from '@multiformats/multiaddr';
import { codes } from '../errors.js';
import { streamToMaConnection } from '@libp2p/utils/stream-to-ma-conn';
import { RELAY_V2_HOP_CODEC, RELAY_V2_STOP_CODEC } from './multicodec.js';
import { createListener } from './listener.js';
import { symbol } from '@libp2p/interface-transport';
import { peerIdFromString } from '@libp2p/peer-id';
import { handleHopProtocol } from './hop.js';
import { handleStop } from './stop.js';
import { pbStream } from 'it-pb-stream';
import pDefer from 'p-defer';
const log = logger('libp2p:circuit');
export class Circuit {
    constructor(components, options) {
        this.components = components;
        this._init = options;
        this.reservationStore = new ReservationStore({
            defaultDataLimit: options.hop?.limit?.data,
            defaultDurationLimit: options.hop?.limit?.duration
        });
        this._started = false;
    }
    isStarted() {
        return this._started;
    }
    async start() {
        if (this._started) {
            return;
        }
        this._started = true;
        // only handle hop if enabled
        if (this._init.hop.enabled === true) {
            void this.components.registrar.handle(RELAY_V2_HOP_CODEC, (data) => {
                void this.onHop(data).catch(err => {
                    log.error(err);
                });
            })
                .catch(err => {
                log.error(err);
            });
        }
        void this.components.registrar.handle(RELAY_V2_STOP_CODEC, (data) => {
            void this.onStop(data).catch(err => {
                log.error(err);
            });
        })
            .catch(err => {
            log.error(err);
        });
        if (this._init.hop.enabled === true) {
            void this.reservationStore.start();
        }
    }
    async stop() {
        if (this._init.hop.enabled === true) {
            this.reservationStore.stop();
            await this.components.registrar.unhandle(RELAY_V2_HOP_CODEC);
        }
        await this.components.registrar.unhandle(RELAY_V2_STOP_CODEC);
    }
    get [symbol]() {
        return true;
    }
    get [Symbol.toStringTag]() {
        return 'libp2p/circuit-relay-v2';
    }
    async onHop({ connection, stream }) {
        log('received circuit v2 hop protocol stream from %s', connection.remotePeer);
        const hopTimeoutPromise = pDefer();
        const timeout = setTimeout(() => {
            hopTimeoutPromise.reject('timed out');
        }, this._init.hop.timeout);
        const pbstr = pbStream(stream);
        try {
            const request = await Promise.race([
                pbstr.pb(CircuitV2.HopMessage).read(),
                hopTimeoutPromise.promise
            ]);
            if (request?.type == null) {
                throw new Error('request was invalid, could not read from stream');
            }
            await Promise.race([
                handleHopProtocol({
                    connection,
                    stream: pbstr,
                    connectionManager: this.components.connectionManager,
                    relayPeer: this.components.peerId,
                    relayAddrs: this.components.addressManager.getListenAddrs(),
                    reservationStore: this.reservationStore,
                    peerStore: this.components.peerStore,
                    request
                }),
                hopTimeoutPromise.promise
            ]);
        }
        catch (_err) {
            pbstr.pb(CircuitV2.HopMessage).write({
                type: CircuitV2.HopMessage.Type.STATUS,
                status: CircuitV2.Status.MALFORMED_MESSAGE
            });
            stream.abort(_err);
        }
        finally {
            clearTimeout(timeout);
        }
    }
    async onStop({ connection, stream }) {
        const pbstr = pbStream(stream);
        const request = await pbstr.readPB(CircuitV2.StopMessage);
        log('received circuit v2 stop protocol request from %s', connection.remotePeer);
        if (request?.type === undefined) {
            return;
        }
        const mStream = await handleStop({
            connection,
            pbstr,
            request
        });
        if (mStream != null) {
            const remoteAddr = multiaddr(request.peer?.addrs?.[0]);
            const localAddr = this.components.transportManager.getAddrs()[0];
            const maConn = streamToMaConnection({
                stream: mStream,
                remoteAddr,
                localAddr
            });
            log('new inbound connection %s', maConn.remoteAddr);
            const conn = await this.components.upgrader.upgradeInbound(maConn);
            log('%s connection %s upgraded', 'inbound', maConn.remoteAddr);
            this.handler?.(conn);
        }
    }
    /**
     * Dial a peer over a relay
     */
    async dial(ma, options = {}) {
        // Check the multiaddr to see if it contains a relay and a destination peer
        const addrs = ma.toString().split('/p2p-circuit');
        const relayAddr = multiaddr(addrs[0]);
        const destinationAddr = multiaddr(addrs[addrs.length - 1]);
        const relayId = relayAddr.getPeerId();
        const destinationId = destinationAddr.getPeerId();
        if (relayId == null || destinationId == null) {
            const errMsg = 'Circuit relay dial failed as addresses did not have peer id';
            log.error(errMsg);
            throw createError(new Error(errMsg), codes.ERR_RELAYED_DIAL);
        }
        const relayPeer = peerIdFromString(relayId);
        const destinationPeer = peerIdFromString(destinationId);
        let disconnectOnFailure = false;
        const relayConnections = this.components.connectionManager.getConnections(relayPeer);
        let relayConnection = relayConnections[0];
        if (relayConnection == null) {
            await this.components.peerStore.addressBook.add(relayPeer, [relayAddr]);
            relayConnection = await this.components.connectionManager.openConnection(relayPeer, options);
            disconnectOnFailure = true;
        }
        try {
            const stream = await relayConnection.newStream([RELAY_V2_HOP_CODEC]);
            return await this.connectV2({
                stream,
                connection: relayConnection,
                destinationPeer,
                destinationAddr,
                relayAddr,
                ma,
                disconnectOnFailure
            });
        }
        catch (err) {
            log.error('Circuit relay dial failed', err);
            disconnectOnFailure && await relayConnection.close();
            throw err;
        }
    }
    async connectV2({ stream, connection, destinationPeer, destinationAddr, relayAddr, ma, disconnectOnFailure }) {
        try {
            const pbstr = pbStream(stream);
            const hopstr = pbstr.pb(CircuitV2.HopMessage);
            hopstr.write({
                type: CircuitV2.HopMessage.Type.CONNECT,
                peer: {
                    id: destinationPeer.toBytes(),
                    addrs: [multiaddr(destinationAddr).bytes]
                }
            });
            const status = await hopstr.read();
            if (status.status !== CircuitV2.Status.OK) {
                throw createError(new Error(`failed to connect via relay with status ${status?.status?.toString() ?? 'undefined'}`), codes.ERR_HOP_REQUEST_FAILED);
            }
            // TODO: do something with limit and transient connection
            let localAddr = relayAddr;
            localAddr = localAddr.encapsulate(`/p2p-circuit/p2p/${this.components.peerId.toString()}`);
            const maConn = streamToMaConnection({
                stream: pbstr.unwrap(),
                remoteAddr: ma,
                localAddr
            });
            log('new outbound connection %s', maConn.remoteAddr);
            const conn = await this.components.upgrader.upgradeOutbound(maConn);
            return conn;
        }
        catch (err) {
            log.error('Circuit relay dial failed', err);
            disconnectOnFailure && await connection.close();
            throw err;
        }
    }
    /**
     * Create a listener
     */
    createListener(options) {
        // Called on successful HOP and STOP requests
        this.handler = options.handler;
        return createListener({
            connectionManager: this.components.connectionManager,
            peerStore: this.components.peerStore
        });
    }
    /**
     * Filter check for all Multiaddrs that this transport can dial on
     *
     * @param {Multiaddr[]} multiaddrs
     * @returns {Multiaddr[]}
     */
    filter(multiaddrs) {
        multiaddrs = Array.isArray(multiaddrs) ? multiaddrs : [multiaddrs];
        return multiaddrs.filter((ma) => {
            return mafmt.Circuit.matches(ma);
        });
    }
}
//# sourceMappingURL=transport.js.map