import { CID } from 'multiformats/cid';
import { sha256 } from 'multiformats/hashes/sha2';
import { logger } from '@libp2p/logger';
const log = logger('libp2p:circuit:v2:util');
const doRelay = (src, dst) => {
    queueMicrotask(() => {
        void dst.sink(src.source).catch(err => log.error('error while relating streams:', err));
    });
    queueMicrotask(() => {
        void src.sink(dst.source).catch(err => log.error('error while relaying streams:', err));
    });
};
export function createLimitedRelay(source, destination, limit) {
    // trivial case
    if (limit == null) {
        doRelay(source, destination);
        return;
    }
    const dataLimit = limit.data ?? 0n;
    const durationLimit = limit.duration ?? 0;
    const src = durationLimitDuplex(dataLimitDuplex(source, dataLimit), durationLimit);
    const dst = durationLimitDuplex(dataLimitDuplex(destination, dataLimit), durationLimit);
    doRelay(src, dst);
}
const dataLimitSource = (stream, limit) => {
    if (limit === 0n) {
        return stream;
    }
    const source = stream.source;
    stream.source = (async function* () {
        let total = 0n;
        for await (const buf of source) {
            const len = BigInt(buf.byteLength);
            if (total + len > limit) {
                // this is a safe downcast since len is guarantee to be in the range for a number
                const remaining = Number(limit - total);
                try {
                    if (remaining !== 0) {
                        yield buf;
                    }
                }
                finally {
                    stream.abort(new Error('data limit exceeded'));
                }
                return;
            }
            yield buf;
            total += len;
        }
    })();
    return stream;
};
const dataLimitSink = (stream, limit) => {
    if (limit === 0n) {
        return stream;
    }
    const sink = stream.sink;
    stream.sink = async (source) => {
        await sink((async function* () {
            let total = 0n;
            for await (const buf of source) {
                const len = BigInt(buf.byteLength);
                if (total + len > limit) {
                    // this is a safe downcast since len is guarantee to be in the range for a number
                    const remaining = Number(limit - total);
                    try {
                        if (remaining !== 0) {
                            yield buf.subarray(0, remaining);
                        }
                    }
                    finally {
                        stream.abort(new Error('data limit exceeded'));
                    }
                    return;
                }
                total += len;
                yield buf;
            }
        })());
    };
    return stream;
};
const dataLimitDuplex = (stream, limit) => {
    dataLimitSource(stream, limit);
    dataLimitSink(stream, limit);
    return stream;
};
const durationLimitDuplex = (stream, limit) => {
    if (limit === 0) {
        return stream;
    }
    let timedOut = false;
    const timeout = setTimeout(() => {
        timedOut = true;
        stream.abort(new Error('exceeded connection duration limit'));
    }, limit);
    const source = stream.source;
    stream.source = (async function* () {
        try {
            for await (const buf of source) {
                if (timedOut) {
                    return;
                }
                yield buf;
            }
        }
        finally {
            clearTimeout(timeout);
        }
    })();
    return stream;
};
/**
 * Convert a namespace string into a cid
 */
export async function namespaceToCid(namespace) {
    const bytes = new TextEncoder().encode(namespace);
    const hash = await sha256.digest(bytes);
    return CID.createV0(hash);
}
/** returns number of ms beween now and expiration time */
export function getExpiration(expireTime) {
    return Number(expireTime) - new Date().getTime();
}
//# sourceMappingURL=utils.js.map